from typing import List

import numpy as np

from a2t.legacy.relation_classification.mnli import (
    NLIRelationClassifierWithMappingHead,
)

TACRED_LABELS = [
    "no_relation",
    "org:alternate_names",
    "org:city_of_headquarters",
    "org:country_of_headquarters",
    "org:dissolved",
    "org:founded",
    "org:founded_by",
    "org:member_of",
    "org:members",
    "org:number_of_employees/members",
    "org:parents",
    "org:political/religious_affiliation",
    "org:shareholders",
    "org:stateorprovince_of_headquarters",
    "org:subsidiaries",
    "org:top_members/employees",
    "org:website",
    "per:age",
    "per:alternate_names",
    "per:cause_of_death",
    "per:charges",
    "per:children",
    "per:cities_of_residence",
    "per:city_of_birth",
    "per:city_of_death",
    "per:countries_of_residence",
    "per:country_of_birth",
    "per:country_of_death",
    "per:date_of_birth",
    "per:date_of_death",
    "per:employee_of",
    "per:origin",
    "per:other_family",
    "per:parents",
    "per:religion",
    "per:schools_attended",
    "per:siblings",
    "per:spouse",
    "per:stateorprovince_of_birth",
    "per:stateorprovince_of_death",
    "per:stateorprovinces_of_residence",
    "per:title",
]

TACRED_BASIC_LABELS = [
    "no_relation",
    "org:dissolved",
    "org:founded",
    "org:founded_by",
    "org:member_of",
    "org:members",
    "org:number_of_employees/members",
    "org:parents",
    "org:political/religious_affiliation",
    "org:shareholders",
    "org:subsidiaries",
    "org:top_members/employees",
    "org:website",
    "per:age",
    "per:cause_of_death",
    "per:charges",
    "per:children",
    "per:date_of_birth",
    "per:date_of_death",
    "per:employee_of",
    "per:other_family",
    "per:parents",
    "per:religion",
    "per:schools_attended",
    "per:siblings",
    "per:spouse",
    "per:title",
    "per:place_of_birth",
    "per:place_of_death",
    "per:place_of_residence",
    "org:place_of_headquarters",
    "both:alternate_names",
]

TACRED_BASIC_LABELS_MAPPING = {
    # Place of birth
    "per:city_of_birth": "per:place_of_birth",
    "per:country_of_birth": "per:place_of_birth",
    "per:stateorprovince_of_birth": "per:place_of_birth",
    # Origin can be see as a country/place of birth
    "per:origin": "per:place_of_birth",
    # Place of death
    "per:city_of_death": "per:place_of_death",
    "per:country_of_death": "per:place_of_death",
    "per:stateorprovince_of_death": "per:place_of_death",
    # Place of residence
    "per:cities_of_residence": "per:place_of_residence",
    "per:countries_of_residence": "per:place_of_residence",
    "per:stateorprovinces_of_residence": "per:place_of_residence",
    # Place of headquarters
    "org:city_of_headquarters": "org:place_of_headquarters",
    "org:country_of_headquarters": "org:place_of_headquarters",
    "org:stateorprovince_of_headquarters": "org:place_of_headquarters",
    # Alternative names
    "per:alternate_names": "both:alternate_names",
    "org:alternate_names": "both:alternate_names",
}

TACRED_LABEL_TEMPLATES = {
    "no_relation": ["{subj} and {obj} are not related"],
    "per:alternate_names": ["{subj} is also known as {obj}"],
    "per:date_of_birth": [
        "{subj}\u00e2\u0080\u0099s birthday is on {obj}",
        "{subj} was born in {obj}",
    ],
    "per:age": ["{subj} is {obj} years old"],
    "per:country_of_birth": ["{subj} was born in {obj}"],
    "per:stateorprovince_of_birth": ["{subj} was born in {obj}"],
    "per:city_of_birth": ["{subj} was born in {obj}"],
    "per:origin": ["{obj} is the nationality of {subj}"],
    "per:date_of_death": ["{subj} died in {obj}"],
    "per:country_of_death": ["{subj} died in {obj}"],
    "per:stateorprovince_of_death": ["{subj} died in {obj}"],
    "per:city_of_death": ["{subj} died in {obj}"],
    "per:cause_of_death": ["{obj} is the cause of {subj}’s death"],
    "per:countries_of_residence": [
        "{subj} lives in {obj}",
        "{subj} has a legal order to stay in {obj}",
    ],
    "per:stateorprovinces_of_residence": [
        "{subj} lives in {obj}",
        "{subj} has a legal order to stay in {obj}",
    ],
    "per:cities_of_residence": [
        "{subj} lives in {obj}",
        "{subj} has a legal order to stay in {obj}",
    ],
    "per:schools_attended": [
        "{subj} studied in {obj}",
        "{subj} graduated from {obj}",
    ],
    "per:title": ["{subj} is a {obj}"],
    "per:employee_of": [
        "{subj} is member of {obj}",
        "{subj} is an employee of {obj}",
    ],
    "per:religion": [
        "{subj} belongs to {obj} religion",
        "{obj} is the religion of {subj}",
        "{subj} believe in {obj}",
    ],
    "per:spouse": [
        "{subj} is the spouse of {obj}",
        "{subj} is the wife of {obj}",
        "{subj} is the husband of {obj}",
    ],
    "per:parents": [
        "{obj} is the parent of {subj}",
        "{obj} is the mother of {subj}",
        "{obj} is the father of {subj}",
        "{subj} is the son of {obj}",
        "{subj} is the daughter of {obj}",
    ],
    "per:children": [
        "{subj} is the parent of {obj}",
        "{subj} is the mother of {obj}",
        "{subj} is the father of {obj}",
        "{obj} is the son of {subj}",
        "{obj} is the daughter of {subj}",
    ],
    "per:siblings": [
        "{subj} and {obj} are siblings",
        "{subj} is brother of {obj}",
        "{subj} is sister of {obj}",
    ],
    "per:other_family": [
        "{subj} and {obj} are family",
        "{subj} is a brother in law of {obj}",
        "{subj} is a sister in law of {obj}",
        "{subj} is the cousin of {obj}",
        "{subj} is the uncle of {obj}",
        "{subj} is the aunt of {obj}",
        "{subj} is the grandparent of {obj}",
        "{subj} is the grandmother of {obj}",
        "{subj} is the grandson of {obj}",
        "{subj} is the granddaughter of {obj}",
    ],
    "per:charges": [
        "{subj} was convicted of {obj}",
        "{obj} are the charges of {subj}",
    ],
    "org:alternate_names": ["{subj} is also known as {obj}"],
    "org:political/religious_affiliation": [
        "{subj} has political affiliation with {obj}",
        "{subj} has religious affiliation with {obj}",
    ],
    "org:top_members/employees": [
        "{obj} is a high level member of {subj}",
        "{obj} is chairman of {subj}",
        "{obj} is president of {subj}",
        "{obj} is director of {subj}",
    ],
    "org:number_of_employees/members": [
        "{subj} employs nearly {obj} people",
        "{subj} has about {obj} employees",
    ],
    "org:members": ["{obj} is member of {subj}", "{obj} joined {subj}"],
    "org:member_of": ["{subj} is member of {obj}", "{subj} joined {obj}"],
    "org:subsidiaries": [
        "{obj} is a subsidiary of {subj}",
        "{obj} is a branch of {subj}",
    ],
    "org:parents": [
        "{subj} is a subsidiary of {obj}",
        "{subj} is a branch of {obj}",
    ],
    "org:founded_by": ["{subj} was founded by {obj}", "{obj} founded {subj}"],
    "org:founded": [
        "{subj} was founded in {obj}",
        "{subj} was formed in {obj}",
    ],
    "org:dissolved": [
        "{subj} existed until {obj}",
        "{subj} disbanded in {obj}",
        "{subj} dissolved in {obj}",
    ],
    "org:country_of_headquarters": [
        "{subj} has its headquarters in {obj}",
        "{subj} is located in {obj}",
    ],
    "org:stateorprovince_of_headquarters": [
        "{subj} has its headquarters in {obj}",
        "{subj} is located in {obj}",
    ],
    "org:city_of_headquarters": [
        "{subj} has its headquarters in {obj}",
        "{subj} is located in {obj}",
    ],
    "org:shareholders": ["{obj} holds shares in {subj}"],
    "org:website": [
        "{obj} is the URL of {subj}",
        "{obj} is the website of {subj}",
    ],
}

TACRED_VALID_CONDITIONS = {
    "per:alternate_names": ["PERSON:PERSON", "PERSON:MISC"],
    "per:date_of_birth": ["PERSON:DATE"],
    "per:age": ["PERSON:NUMBER", "PERSON:DURATION"],
    "per:country_of_birth": ["PERSON:COUNTRY"],
    "per:stateorprovince_of_birth": ["PERSON:STATE_OR_PROVINCE"],
    "per:city_of_birth": ["PERSON:CITY"],
    "per:origin": ["PERSON:NATIONALITY", "PERSON:COUNTRY", "PERSON:LOCATION"],
    "per:date_of_death": ["PERSON:DATE"],
    "per:country_of_death": ["PERSON:COUNTRY"],
    "per:stateorprovince_of_death": ["PERSON:STATE_OR_PROVICE"],
    "per:city_of_death": ["PERSON:CITY", "PERSON:LOCATION"],
    "per:cause_of_death": ["PERSON:CAUSE_OF_DEATH"],
    "per:countries_of_residence": ["PERSON:COUNTRY", "PERSON:NATIONALITY"],
    "per:stateorprovinces_of_residence": ["PERSON:STATE_OR_PROVINCE"],
    "per:cities_of_residence": ["PERSON:CITY", "PERSON:LOCATION"],
    "per:schools_attended": ["PERSON:ORGANIZATION"],
    "per:title": ["PERSON:TITLE"],
    "per:employee_of": ["PERSON:ORGANIZATION"],
    "per:religion": ["PERSON:RELIGION"],
    "per:spouse": ["PERSON:PERSON"],
    "per:parents": ["PERSON:PERSON"],
    "per:children": ["PERSON:PERSON"],
    "per:siblings": ["PERSON:PERSON"],
    "per:other_family": ["PERSON:PERSON"],
    "per:charges": ["PERSON:CRIMINAL_CHARGE"],
    "org:alternate_names": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:MISC"],
    "org:political/religious_affiliation": [
        "ORGANIZATION:RELIGION",
        "ORGANIZATION:IDEOLOGY",
    ],
    "org:top_members/employees": ["ORGANIZATION:PERSON"],
    "org:number_of_employees/members": ["ORGANIZATION:NUMBER"],
    "org:members": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:COUNTRY"],
    "org:member_of": [
        "ORGANIZATION:ORGANIZATION",
        "ORGANIZATION:COUNTRY",
        "ORGANIZATION:LOCATION",
        "ORGANIZATION:STATE_OR_PROVINCE",
    ],
    "org:subsidiaries": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:LOCATION"],
    "org:parents": ["ORGANIZATION:ORGANIZATION", "ORGANIZATION:COUNTRY"],
    "org:founded_by": ["ORGANIZATION:PERSON"],
    "org:founded": ["ORGANIZATION:DATE"],
    "org:dissolved": ["ORGANIZATION:DATE"],
    "org:country_of_headquarters": ["ORGANIZATION:COUNTRY"],
    "org:stateorprovince_of_headquarters": ["ORGANIZATION:STATE_OR_PROVINCE"],
    "org:city_of_headquarters": ["ORGANIZATION:CITY", "ORGANIZATION:LOCATION"],
    "org:shareholders": ["ORGANIZATION:PERSON", "ORGANIZATION:ORGANIZATION"],
    "org:website": ["ORGANIZATION:URL"],
}


class TACREDClassifier(NLIRelationClassifierWithMappingHead):
    def __init__(self, **kwargs):
        pretrained_model = kwargs.pop("pretrained_model", "microsoft/deberta-v2-xlarge-mnli")
        super(TACREDClassifier, self).__init__(
            pretrained_model=pretrained_model,
            labels=TACRED_LABELS,
            template_mapping=TACRED_LABEL_TEMPLATES,
            valid_conditions=TACRED_VALID_CONDITIONS,
            entailment_position=2,
            **kwargs
        )



import argparse
import json
import os
from pprint import pprint

import numpy as np
from sklearn.metrics import precision_recall_fscore_support

from a2t.legacy.topic_classification.mlm import MLMTopicClassifier
from a2t.legacy.topic_classification.mnli import (
    NLITopicClassifier,
    NLITopicClassifierWithMappingHead,
)
from a2t.legacy.topic_classification.nsp import NSPTopicClassifier

CLASSIFIERS = {
    "mnli": NLITopicClassifier,
    "nsp": NSPTopicClassifier,
    "mlm": MLMTopicClassifier,
    "mnli-mapping": NLITopicClassifierWithMappingHead,
}


def top_k_accuracy(output, labels, k=5):
    preds = np.argsort(output)[:, ::-1][:, :k]
    return sum(l in p for l, p in zip(labels, preds)) / len(labels)


parser = argparse.ArgumentParser(
    prog="run_evaluation",
    description="Run a evaluation for each configuration.",
)
parser.add_argument("dataset", type=str, help="Dataset file.")
parser.add_argument("topics", type=str, help="Topics or classes file.")
parser.add_argument(
    "--config",
    type=str,
    dest="config",
    help="Configuration file for the experiment.",
)

args = parser.parse_args()

with open(args.topics, "rt") as f:
    topics = [topic.rstrip().replace("_", " ") for topic in f]

topic2id = {topic: i for i, topic in enumerate(topics)}

with open(args.dataset, "rt") as f:
    contexts, labels = [], []
    for line in f:
        _, label, context = line.strip().split("\t")
        contexts.append(context)
        labels.append(topic2id[label])

labels = np.array(labels)

with open(args.config, "rt") as f:
    config = json.load(f)

for configuration in config:
    os.makedirs(f"experiments/{configuration['name']}", exist_ok=True)
    classifier = CLASSIFIERS[configuration["classification_model"]](labels=topics, **configuration)
    output = classifier(contexts, batch_size=configuration["batch_size"])
    np.save(f"experiments/{configuration['name']}/output.npy", output)
    np.save(f"experiments/{configuration['name']}/labels.npy", labels)
    pre, rec, f1, _ = precision_recall_fscore_support(labels, np.argmax(output, -1), average="weighted")
    configuration["precision"] = pre
    configuration["recall"] = rec
    configuration["f1-score"] = f1
    configuration["top-1"] = top_k_accuracy(output, labels, k=1)
    configuration["top-3"] = top_k_accuracy(output, labels, k=3)
    configuration["top-5"] = top_k_accuracy(output, labels, k=5)
    configuration["topk-curve"] = [top_k_accuracy(output, labels, k=i) for i in range(len(topics))]
    pprint(configuration)
    print(pre, rec, f1, top_k_accuracy(output, labels, k=1))

with open(args.config, "wt") as f:
    json.dump(config, f, indent=4)



import argparse
import gzip
import json
import os

import numpy as np

from a2t.legacy.topic_classification.mlm import MLMTopicClassifier
from a2t.legacy.topic_classification.mnli import (
    NLITopicClassifier,
    NLITopicClassifierWithMappingHead,
)
from a2t.legacy.topic_classification.nsp import NSPTopicClassifier

CLASSIFIERS = {
    "mnli": NLITopicClassifier,
    "nsp": NSPTopicClassifier,
    "mlm": MLMTopicClassifier,
    "mnli-mapping": NLITopicClassifierWithMappingHead,
}

parser = argparse.ArgumentParser(
    prog="annotate_wordnet",
    description="Annotate WordNet glosses with domain information.",
)
parser.add_argument("dataset", type=str, help="Dataset file.")
parser.add_argument("topics", type=str, help="Topics or classes file.")
parser.add_argument(
    "--config",
    type=str,
    dest="config",
    help="Configuration file for the experiment.",
)

args = parser.parse_args()

with open(args.topics, "rt") as f:
    topics = [topic.rstrip().replace("_", " ") for topic in f]

topic2id = {topic: i for i, topic in enumerate(topics)}

senses, glosses = [], []
with gzip.open(args.dataset, "rt") as f:
    for line in f:
        row = line.split()
        idx, gloss = row[0], " ".join(row[1:])
        gloss = gloss.replace("\t", " ")
        senses.append(idx)
        glosses.append(gloss)

with open(args.config, "rt") as f:
    config = json.load(f)

for configuration in config:
    os.makedirs(f"experiments/{configuration['name']}", exist_ok=True)
    classifier = CLASSIFIERS[configuration["classification_model"]](labels=topics, **configuration)
    output = classifier(glosses, batch_size=configuration["batch_size"])
    # Normalize the output to avoid low confidence level due to large ammount of labels
    output = (output - output.min()) / (output.max() - output.min())
    labels = [topics[i] for i in np.argmax(output, -1)]
    confidences = [x for x in np.max(output, -1)]

    output_name = os.path.basename(args.dataset).replace("list", "annotations")

    with gzip.open(f"experiments/{configuration['name']}/{output_name}", "wt") as out_f:
        for sense, label, confidence in zip(senses, labels, confidences):
            out_f.write(f"{sense}\t{label}\t{confidence}\n")



import sys
from collections import defaultdict
from pprint import pprint
from typing import Dict, List

import numpy as np
import torch
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from a2t.base import Classifier, np_softmax


class _NLITopicClassifier(Classifier):
    def __init__(
        self,
        labels: List[str],
        *args,
        pretrained_model: str = "roberta-large-mnli",
        use_cuda=True,
        query_phrase="The domain of the sentence is about",
        entailment_position=2,
        half=False,
        verbose=True,
        **kwargs,
    ):
        super().__init__(
            labels,
            pretrained_model=pretrained_model,
            use_cuda=use_cuda,
            verbose=verbose,
            half=half,
        )
        self.query_phrase = query_phrase
        self.ent_pos = entailment_position

        def idx2topic(idx):
            return self.labels[idx]

        self.idx2topic = np.vectorize(idx2topic)

    def _initialize(self, pretrained_model):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.model = AutoModelForSequenceClassification.from_pretrained(pretrained_model)

    def _run_batch(self, batch):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True)
            input_ids = torch.tensor(input_ids["input_ids"]).to(self.device)
            output = self.model(input_ids)[0][:, self.ent_pos].view(input_ids.shape[0] // len(self.labels), -1)
            output = output.detach().cpu().numpy()

        return output

    def __call__(self, contexts: List[str], batch_size: int = 1):
        if not isinstance(contexts, list):
            contexts = [contexts]

        batch, outputs = [], []
        for i, context in tqdm(enumerate(contexts), total=len(contexts)):
            sentences = [f"{context} {self.tokenizer.sep_token} {self.query_phrase} {topic}." for topic in self.labels]
            batch.extend(sentences)

            if (i + 1) % batch_size == 0:
                output = self._run_batch(batch)
                outputs.append(output)
                batch = []

        if len(batch) > 0:
            output = self._run_batch(batch)
            outputs.append(output)

        outputs = np.vstack(outputs)

        return outputs

    def predict(
        self,
        contexts: List[str],
        batch_size: int = 1,
        return_labels: bool = True,
        return_confidences: bool = False,
        topk: int = 1,
    ):
        output = self(contexts, batch_size)
        topics = np.argsort(output, -1)[:, ::-1][:, :topk]
        if return_labels:
            topics = self.idx2topic(topics)
        if return_confidences:
            topics = np.stack((topics, np.sort(output, -1)[:, ::-1][:, :topk]), -1).tolist()
            topics = [
                [(int(label), float(conf)) if not return_labels else (label, float(conf)) for label, conf in row]
                for row in topics
            ]
        else:
            topics = topics.tolist()
        if topk == 1:
            topics = [row[0] for row in topics]

        return topics


class NLITopicClassifier(_NLITopicClassifier):
    """NLITopicClassifier

    Zero-Shot topic classifier based on a Natural Language Inference pretrained Transformer.

    Use:

    ```python
    >>> from a2t.topic_classification import NLITopicClassifier

    >>> topics = ['politics', 'culture', 'economy', 'biology', 'legal', 'medicine', 'business']
    >>> context = "hospital: a health facility where patients receive treatment."

    >>> clf = NLITopicClassifier(topics)

    >>> predictions = clf(context)[0]
    >>> print(sorted(list(zip(predictions, topics)), reverse=True))

    [(0.77885467, 'medicine'),
     (0.08395168, 'biology'),
     (0.040319894, 'business'),
     (0.027866213, 'economy'),
     (0.02357693, 'politics'),
     (0.023382403, 'legal'),
     (0.02204825, 'culture')]

    ```
    """

    def __init__(
        self,
        labels: List[str],
        *args,
        pretrained_model: str = "roberta-large-mnli",
        **kwargs,
    ):
        super(NLITopicClassifier, self).__init__(labels, *args, pretrained_model=pretrained_model, **kwargs)

    def __call__(self, contexts: List[str], batch_size: int = 1):
        outputs = super().__call__(contexts=contexts, batch_size=batch_size)
        outputs = np_softmax(outputs)

        return outputs


class NLITopicClassifierWithMappingHead(_NLITopicClassifier):
    def __init__(
        self,
        labels: List[str],
        topic_mapping: Dict[str, str],
        pretrained_model: str = "roberta-large-mnli",
        *args,
        **kwargs,
    ):
        self.new_topics = list(topic_mapping.keys())
        self.target_topics = labels
        self.new_topics2id = {t: i for i, t in enumerate(self.new_topics)}
        self.mapping = defaultdict(list)
        for key, value in topic_mapping.items():
            self.mapping[value].append(self.new_topics2id[key])

        super().__init__(self.new_topics, *args, pretrained_model=pretrained_model, **kwargs)

        def idx2topic(idx):
            return self.target_topics[idx]

        self.idx2topic = np.vectorize(idx2topic)

    def __call__(self, contexts, batch_size=1):
        outputs = super().__call__(contexts, batch_size)
        outputs = np.hstack([np.max(outputs[:, self.mapping[topic]], axis=-1, keepdims=True) for topic in self.target_topics])
        outputs = np_softmax(outputs)

        return outputs


if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("Usage:\tpython3 get_topics.py topics.txt input_file.txt\n\tpython3 get_topics.py topics.txt < " "input_file.txt")
        exit(1)

    with open(sys.argv[1], "rt") as f:
        topics = [topic.rstrip().replace("_", " ") for topic in f]

    input_stream = open(sys.argv[2], "rt") if len(sys.argv) == 3 else sys.stdin

    clf = NLITopicClassifier(labels=topics, pretrained_model="roberta-large-mnli")

    for line in input_stream:
        line = line.rstrip()
        output_probs = clf(line)[0]
        topic_dist = sorted(list(zip(output_probs, topics)), reverse=True)
        print(line)
        pprint(topic_dist)
        print()
